#include "dtc.h"

#define ONE_OVER_SQRT_3 (d_t)0.577350269
#define ONE_OVER_THREE (d_t)0.333333333
#define TWO_OVER_THREE (d_t)0.66666666

void ComputeAbcToAlphaBeta(d_t a, d_t b, d_t c, d_t &alpha, d_t &beta);

void ComputeRectangularToPolar(d_t x, d_t y, d_t &r, d_t &theta);

d_t ComputeTorque(d_t a, d_t b, d_t c, d_t d, d_t p);

ap_uint<2> ComputeTemHystState(ap_uint<2> previous_state, d_t Tem_s, d_t ref_Tem, d_t epsilon_Tem);

ap_uint<1> ComputePhiHystState(ap_uint<1> previous_state, d_t phi_s, d_t ref_phi, d_t epsilon_phi);

ap_uint<3> ComputeSector(d_t theta);

static ap_uint<3> pwm_lut[2][3][6] =
{
	{
		{0b100, 0b101, 0b001, 0b011, 0b010, 0b110},
		{0b000, 0b111, 0b000, 0b111, 0b000, 0b111},
		{0b010, 0b110, 0b100, 0b101, 0b001, 0b011}
	},
	{
		{0b101, 0b001, 0b011, 0b010, 0b110, 0b100},
		{0b111, 0b000, 0b111, 0b000, 0b111, 0b000},
		{0b011, 0b010, 0b110, 0b100, 0b101, 0b001}
	}
};
#ifdef TESTBENCH
ap_uint<3> dtc(inputs &ins, outputs &outs)
#else
ap_uint<3> dtc(inputs &ins)
#endif
{
	d_t adc0 = ins.adc0 + ins.adc_offset0;
	d_t adc1 = ins.adc1 + ins.adc_offset1;
	d_t adc2 = ins.adc2 + ins.adc_offset2;

	// ------------------------------------------------------------------------

	d_t i_alpha, i_beta;

	ComputeAbcToAlphaBeta(ins.adc0, ins.adc1, ins.adc2, i_alpha, i_beta);

	// ------------------------------------------------------------------------

	d_t Ld = ((ap_fixed<W,I>)ins.l / 32768);

	d_t i_alpha_times_l = i_alpha * Ld;
	d_t i_beta_times_l  = i_beta * Ld;

	d_t psi_alpha = i_alpha_times_l + ins.phi_alpha_r;
	d_t psi_beta  = i_beta_times_l  + ins.phi_beta_r;

	// ------------------------------------------------------------------------

	d_t phi_s;
	d_t theta_s;

	ComputeRectangularToPolar(psi_alpha, psi_beta, phi_s, theta_s);

	// ------------------------------------------------------------------------

	d_t Tem_s;

	Tem_s = ComputeTorque(psi_alpha, psi_beta, i_alpha, i_beta, ins.p);

	// ------------------------------------------------------------------------

	static ap_uint<2> Tem_hyst_state = 0;

	Tem_hyst_state = ComputeTemHystState(Tem_hyst_state, Tem_s, ins.Tem_ref, ins.epsilon_Tem);

	// ------------------------------------------------------------------------

	static ap_uint<1> phi_hyst_state = 0;

	phi_hyst_state = ComputePhiHystState(phi_hyst_state, phi_s, ins.phi_ref, ins.epsilon_phi);

	// ------------------------------------------------------------------------

	ap_uint<3> sector = ComputeSector(theta_s);

	// ------------------------------------------------------------------------
#ifdef TESTBENCH

	outs.debug_i_alpha = i_alpha;
	outs.debug_i_beta = i_beta;
	outs.debug_psi_alpha = psi_alpha;
	outs.debug_psi_beta = psi_beta;
	outs.debug_phi_s = phi_s;
	outs.debug_theta_s = theta_s * 8;
	outs.debug_Tem_s = Tem_s;
	outs.debug_TemHystState = Tem_hyst_state;
	outs.debug_PhiHystState = phi_hyst_state;
	outs.debug_sector = sector;

#endif


	// ------------------------------------------------------------------------

	return pwm_lut[phi_hyst_state][Tem_hyst_state][sector];
}

void ComputeAbcToAlphaBeta(d_t a, d_t b, d_t c, d_t &alpha, d_t &beta)
{
	alpha = a*ONE_OVER_THREE - b*TWO_OVER_THREE - c*TWO_OVER_THREE;
	beta  = (b - c)*ONE_OVER_SQRT_3;
}

void ComputeRectangularToPolar(d_t x, d_t y, d_t &r, d_t &theta)
{
	ap_fixed<W*2,I*2, AP_TRN, AP_SAT> x_squared = x*x;
	ap_fixed<W*2,I*2, AP_TRN, AP_SAT> y_squared = y*y;

	r = hls::sqrt((ap_fixed<64,32>)(x_squared + y_squared));

	#pragma HLS PIPELINE
	theta = hls::atan2((ap_fixed<W,I>)y, (ap_fixed<W,I>)x);
}

d_t ComputeTorque(d_t a, d_t b, d_t c, d_t d, d_t p)
{
	ap_fixed<W*2, I*2> a_times_d = a*d;
	ap_fixed<W*2, I*2> b_times_c = b*c;

	ap_fixed<W, I> temp = (ap_fixed<W*2, I*2>)(a_times_d - b_times_c) / 4096;

	return (d_t)(1.5) * p * temp;
}

ap_uint<2> ComputeTemHystState(ap_uint<2> previous_state, d_t Tem_s, d_t ref_Tem, d_t epsilon_Tem)
{
	d_t ref_Tem_minus_Tem_s = ref_Tem - Tem_s;

	if(previous_state == 0)
	{
		if(ref_Tem_minus_Tem_s >  epsilon_Tem)
		{
			return 2;
		}
		else if(ref_Tem_minus_Tem_s >  0)
		{
			return 1;
		}
	}
	else if(previous_state == 1)
	{
		if(ref_Tem_minus_Tem_s >  epsilon_Tem)
		{
			return 2;
		}
		else if(ref_Tem_minus_Tem_s < - epsilon_Tem)
		{
			return 0;
		}
	}
	else
	{
		if(ref_Tem_minus_Tem_s < - epsilon_Tem)
		{
			return 0;
		}
		else if(ref_Tem_minus_Tem_s < 0)
		{
			return 1;
		}
	}

	return previous_state;
}

ap_uint<1> ComputePhiHystState(ap_uint<1> previous_state, d_t phi_s, d_t ref_phi, d_t epsilon_phi)
{
	d_t ref_phi_minus_phi_s = ref_phi - phi_s;

	if(previous_state == 0)
	{
		if(ref_phi_minus_phi_s >  epsilon_phi)
		{
			return 1;
		}
	}
	else
	{
		if(ref_phi_minus_phi_s < - epsilon_phi)
		{
			return 0;
		}
	}

	return previous_state;
}

ap_uint<3> ComputeSector(d_t theta)
{
	if(theta < 0)
		theta += (d_t)PI_FIXED * (d_t)2;

	if(theta < (d_t)PI_FIXED/(d_t)6)
		return 0;
	else if(theta < (d_t)PI_FIXED/(d_t)2)
		return 1;
	else if(theta < (d_t)5*(d_t)PI_FIXED/(d_t)6)
		return 2;
	else if(theta < (d_t)7*(d_t)PI_FIXED/(d_t)6)
		return 3;
	else if(theta < (d_t)3*(d_t)PI_FIXED/(d_t)2)
		return 4;
	else if(theta < (d_t)11*(d_t)PI_FIXED/(d_t)6)
		return 5;
	else if(theta < (d_t)2*(d_t)PI_FIXED)
		return 0;

	return 0;
}
